import java.awt.Color;
import java.awt.geom.Ellipse2D;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.util.List;


public class HumanHeadDetectorCLI {
	
	
	static String getOutputFilename(String prefix) {
		return(prefix + "_HFD.png");
	}
	
	
	public static void main(String[] args) {

		if(args.length<2) {
			System.out.println("usage: HumanFaceDetectorCLI [imagefile] [minMinorAxis] [maxMinorAxis] [no_ellipses_to_detect]");
			return;
		}
		
		// open input image
		MImage input = null;
		try {
			input = new MImage(new File(args[0]));
		} catch (IOException e1) {
			e1.printStackTrace();
		}

		// legge i parametri
		int minMinorAxis = 10;
		if(args.length>1)
		{
			try {
				minMinorAxis = Integer.parseInt(args[1]);
			} catch (NumberFormatException e) {
		        System.err.println("Invalid minMinorAxis number. Assuming 10.");
		    }

		}
		int maxMinorAxis = 50;
		if(args.length>2)
		{
			try {
				maxMinorAxis = Integer.parseInt(args[2]);
			} catch (NumberFormatException e) {
		        System.err.println("Invalid maxMinorAxis number. Assuming 50.");
		    }

		}
		int no_ellipses = 1;
		if(args.length>3)
		{
			try {
				no_ellipses = Integer.parseInt(args[3]);
			} catch (NumberFormatException e) {
		        System.err.println("Invalid ellipses number. Assuming 1.");
		    }

		}
		System.out.println("parameters: minMinorAxis=" + minMinorAxis + ", maxMinorAxis=" + maxMinorAxis + ", no_ellipses=" + no_ellipses);
		
		// process the input image
		System.out.println("processing...");

		// stage 0: caricamento dell'immagine di input
		MImage c;
		try {
			c = (MImage) input.clone();
		} catch (CloneNotSupportedException e) {
			c = input;
		}
		
		// stage 1: preprocessing
		c.toRGB();
		c.gaussianBlur();
		c.filter(new HumanSkinFilter());
		c.edge2();
		c.threshold(10);
		c.toBinary();
				
		MImage output = new MImage( c.i.getWidth(), c.i.getHeight(), BufferedImage.TYPE_INT_RGB, Color.BLACK );

		// stage 2: riconoscimento ellissi
		RHTEllipse he = new RHTEllipse( (int)(minMinorAxis*1.5), // minMajorAxis
				                        (int)(minMinorAxis*1.5*1.5), // maxMajorAxis=maxMinorAxis*1.5? 
				                          minMinorAxis,
				                          maxMinorAxis,
				                          10000, // idleStop
				                          no_ellipses,
				                          c.i.getData(),
				                          output.i.getRaster()
				                          );

		he.execute();
		
		List<Ellipse2D.Double> ellipses;
		try {
			ellipses = he.get();
		} catch (Exception e) {
			e.printStackTrace();
			return;
		}
				
		// disegna tutti gli ellissi riconosciuti
		for(int j=0;j<ellipses.size();j++) {
			output.draw(ellipses.get(j), Color.MAGENTA, null);
			input.draw(ellipses.get(j), Color.MAGENTA, null);
		}
		
		// save output image
		try {
			output.save(new File(getOutputFilename(args[0])));
		} catch (IOException e) {
			e.printStackTrace();
		}
		
	}

}
