import java.awt.*;
import java.awt.event.*;
import java.awt.geom.Ellipse2D;
import java.awt.image.BufferedImage;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;
import java.io.FilenameFilter;
import java.io.IOException;
import java.util.List;

import javax.swing.ProgressMonitor;


class HumanHeadDetectorGUI extends Frame {
	
	static Frame mainFrame;
		
	MImage stage0MImage;
	MImage stage1MImage;
	MImage stage2MImage;
	
	MImageCanvas stage0MImageCanvas;
	MImageCanvas stage1MImageCanvas;
	MImageCanvas stage2MImageCanvas;
	
	Button loadButton;
	Button stage1Button;
	Button stage2Button;
	
	private void updateComponentsState() {
		stage0MImageCanvas.set(stage0MImage);
		stage0MImageCanvas.repaint();
		stage1MImageCanvas.set(stage1MImage);
		stage1MImageCanvas.repaint();
		stage2MImageCanvas.set(stage2MImage);
		stage2MImageCanvas.repaint();
		mainFrame.repaint();
	}
	
	HumanHeadDetectorGUI()
	{
		mainFrame = this;
		this.setTitle("HumanHeadDetectorGUI");
		this.addWindowListener( new WindowAdapter() {
			public void windowClosing( WindowEvent e ) {
				System.exit(0);
			}
		});
		
		loadButton = new Button("Load...");
		loadButton.setEnabled(true);
		loadButton.addActionListener( new ActionListener() {
			public void actionPerformed( ActionEvent e ) {
				FileDialog openDialog =
					new FileDialog( mainFrame, "Open", FileDialog.LOAD );
				openDialog.setFilenameFilter( new FilenameFilter() {
					// MEMO: Filename filters do not function in Sun's reference implementation for Microsoft Windows.
					@Override
					public boolean accept(File directory, String filename) {
						if (filename.endsWith(".png")) return true;
						if (filename.endsWith(".gif")) return true;
						if (filename.endsWith(".jpg")) return true;
						if (filename.endsWith(".bmp")) return true;
						return false;
					}
				});
				openDialog.setLocationRelativeTo(mainFrame);
				openDialog.setVisible(true);
				
				if(openDialog.getFile()==null) return; // l'utente ha premuto "Cancel"

				File inputFile = new File( openDialog.getDirectory() + openDialog.getFile() );
				
				try {
					stage0MImage = new MImage(inputFile);
				} catch (IOException e1) {
					new MessageDialog(mainFrame, "Error loading image", e1.getMessage(),true);
					stage0MImage = null;
					stage1MImage = null;
					stage2MImage = null;
					updateComponentsState();
					stage1Button.setEnabled(false);
					stage2Button.setEnabled(false);
					return;
				}
				// nessuna eccezione -> caricamento ok
				mainFrame.setTitle("HumanHeadDetectorGUI - " + inputFile);				
				stage1MImage = null;
				stage2MImage = null;
				updateComponentsState();
				stage1Button.setEnabled(true);
			}
		});
		
		stage1Button = new Button("Preprocess...");
		stage1Button.setEnabled(false);
		stage1Button.addActionListener( new ActionListener() {
			public void actionPerformed( ActionEvent e ) {
				// mostra una finestra per settare le opzioni
				Stage1OptionsDialog options = new Stage1OptionsDialog(mainFrame);
				if(options.isConfirmed == false) return;
				
				try {
					stage1MImage = (MImage) stage0MImage.clone();
				} catch (CloneNotSupportedException e1) {
					stage1MImage = stage0MImage;
				}
				
				// se l'area dell'immagine è > di 1000*1000
				// la ridimensiona automaticamente per velocizzare il riconoscimento
				/*
				if(stage1MImage.getWidth()*stage1MImage.getHeight() > (1000*1000)) {
					stage1MImage.scale(Math.max(stage1MImage.getWidth(),
							                    stage1MImage.getHeight())/1000); // 2FIX: out of memory with big images
				}*/
				
				mainFrame.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
				// ALT: show "processing" window
				
				stage1MImage.toRGB();
				stage1MImage.gaussianBlur();
				if(options.skinSegmentation) // se richiesto, esegue la segmentazione basata su colore
					stage1MImage.filter(new HumanSkinFilter());
				else stage1MImage.toGrayscale();
				//stage1MImage.gaussianBlur();
				//NO? stage1MImage.sharpen();
				stage1MImage.edge2();
				//stage1MImage.edge();
				stage1MImage.threshold(options.getBinaryThreshold());
				stage1MImage.toBinary();
				
				mainFrame.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
				
				stage2MImage = null;
				updateComponentsState();
				stage2Button.setEnabled(true);
			}
		});
		
		stage2Button = new Button("Detect ellipses...");
		stage2Button.setEnabled(false);
		stage2Button.addActionListener( new ActionListener() {
			@Override
		    public void actionPerformed(ActionEvent evt) {
				// mostra una finestra per settare le opzioni
				Stage2OptionsDialog options = new Stage2OptionsDialog(mainFrame);
				if(options.isConfirmed == false) return;
				
				// crea un immagine vuota per l'output
				stage2MImage = new MImage( stage1MImage.getWidth(), stage1MImage.getHeight(), BufferedImage.TYPE_INT_RGB, Color.BLACK );
				
				// crea un progress monitor per mostrare una progress bar (da anche la possibilità di annullare il riconoscimento)
		        final ProgressMonitor progressMonitor;
		    	progressMonitor = new ProgressMonitor(mainFrame, "Ricerca ellissi in corso...", "Ricerca ellissi in corso...", 0, 100);
		        progressMonitor.setProgress(0);
				progressMonitor.setMillisToDecideToPopup(1);
				progressMonitor.setMillisToPopup(1);
		        
				// crea un istanza di RHTEllipse e ne imposta i parametri
		        final RHTEllipse task;
		        task = new RHTEllipse( options.getMinMajorAxis(),
				                        options.getMaxMajorAxis(),
				                        options.getMinMinorAxis(),
				                        options.getMaxMinorAxis(),
				                        10000, // idleStop
				                        options.getNoEllipses(),
				                        stage1MImage.i.getData(),
				                        stage2MImage.i.getRaster()) {
						@Override
						protected void done() { // metodo eseguito al completamento del thread
							super.done();
							if(isCancelled()) {
								//super.done();
								return;
							}
							
	        				// legge il risultato del riconoscimento
	        				List<Ellipse2D.Double> ellipses = null;
	        				try {
	        					//ellipses = task.get();
	        					ellipses = this.get();
	        				} catch (Exception e1) {
	        				}
	        				
	        				if(ellipses==null || ellipses.size()==0) {
	        					new MessageDialog(mainFrame, "Result", "Nessun ellisse trovato con questi parametri!", true);
	        					updateComponentsState();
	        					return;
	        				}
	        				
        					// evidenzia gli ellissi
        					for(int j=0;j<ellipses.size();j++) {
        						//stage0MImage.draw(ellipses.get(j), Color.MAGENTA);
        						//stage1MImage.draw(ellipses.get(j), Color.MAGENTA);
        						stage2MImage.draw(ellipses.get(j), Color.MAGENTA, new BasicStroke(
        					            5,//width
        					            BasicStroke.CAP_SQUARE,
        					            BasicStroke.JOIN_BEVEL,
        					            0.0f,//miterlimit doesn't matter
        					            new float[] {0.1f*72,0.3f*72},//Dash pattern
        					            0.0f)//Dash phase
        							);
        					}
	        				updateComponentsState();
							//super.done();
						}
		        	
		        };
		        
		        // registra un PropertyChangeListener che monitorerà il riconoscimento in background
		        task.addPropertyChangeListener( new PropertyChangeListener() {
		            public void propertyChange(PropertyChangeEvent evt) {
		            	
		                if ("progress" != evt.getPropertyName() ) return;
		                
	                    int progress = (Integer) evt.getNewValue();
	                    progressMonitor.setProgress(progress);
	                    String message =
	                        String.format("Completato %d%%.\n", progress);
	                    progressMonitor.setNote(message);
	                    
                        if (progressMonitor.isCanceled()) {
                            task.cancel(true);
                            task.removePropertyChangeListener(this);
                            stage2MImage = null;
                            return;
                        }
                        
                        /* CANCELLARE: spostato in done()
	                    if (task.isDone()) {
	        				// legge il risultato del riconoscimento
	        				List<Ellipse2D.Double> ellipses = null;
	        				try {
	        					ellipses = task.get();
	        				} catch (Exception e1) {
	        				}
	        				
	        				if(ellipses==null) {
	        					new MessageDialog(mainFrame, "Result", "Nessun ellisse trovato con questi parametri!", true);
	        					updateComponentsState();
	        					return;
	        				}
	        				
        					// evidenzia gli ellissi
        					for(int j=0;j<ellipses.size();j++) {
        						//stage0MImage.draw(ellipses.get(j), Color.MAGENTA);
        						//stage1MImage.draw(ellipses.get(j), Color.MAGENTA);
        						stage2MImage.draw(ellipses.get(j), Color.MAGENTA, new BasicStroke(
        					            5,//width
        					            BasicStroke.CAP_SQUARE,
        					            BasicStroke.JOIN_BEVEL,
        					            0.0f,//miterlimit doesn't matter
        					            new float[] {0.1f*72,0.3f*72},//Dash pattern
        					            0.0f)//Dash phase
        							);
        					}
	        				updateComponentsState();
	                    }*/
		            }
		        });
		        
				//mainFrame.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));

				// avvia il riconoscimento in un thread in background
		        task.execute();
		        
				//mainFrame.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
		        
		        // a questo punto il riconoscimento avviene in un thread in background
		        // il PropertyChangeListener si occuperà del resto...
		    }

		});
		
		
		stage0MImageCanvas = new MImageCanvas();
		stage1MImageCanvas = new MImageCanvas();
		stage2MImageCanvas = new MImageCanvas();
		stage0MImageCanvas.setSize(300, 300);
		stage1MImageCanvas.setSize(300, 300);
		stage2MImageCanvas.setSize(300, 300);

		this.setLayout(new FlowLayout());
		//this.setLayout(new GridLayout(2,3,10,10));
		this.setBounds(0, 0, 950, 400);
	
		this.add(stage0MImageCanvas);
		this.add(stage1MImageCanvas);
		this.add(stage2MImageCanvas);
		this.add(loadButton);
		this.add(stage1Button);
		this.add(stage2Button);
		
		//this.pack();
		this.setVisible(true);
	}
	
	
    public static void main(String[] args) {
    	new HumanHeadDetectorGUI();
    }
	
	
}



// canvas per mostrare le immagini
class MImageCanvas extends Canvas {
    MImage unscaledImage;
    MImage scaledImage;

    public MImageCanvas() {
    	unscaledImage = null;
    	scaledImage = null;
    	
    	// facendo click sul canvas viene mostrata l'immagine originale
    	this.addMouseListener(new MouseAdapter() {
    		public void mouseClicked(MouseEvent e) {
    			if(unscaledImage==null) return; 
    			new MImageFrame(unscaledImage);
    		}
    	});
    }
    
    public void set(MImage i) {
    	this.unscaledImage = i;
    	this.scaledImage = i;
    	if(i==null) return;
    	
    	if(unscaledImage.getWidth()>this.getWidth() || 
    	   unscaledImage.getHeight()>this.getHeight() ) {
			try {
				scaledImage = (MImage) i.clone();
			} catch (CloneNotSupportedException e) {
				return;
			}
			// scale and keep aspect ratio
			scaledImage.scale(Math.min(this.getWidth()/(double)i.getWidth(), this.getHeight()/(double)i.getHeight()));
    	}
    	 
    	//repaint();
    }

    public void paint(Graphics g)
    {      
    	if(scaledImage!=null)
         g.drawImage(scaledImage.getImage(), 0, 0, this);
    	else {
         g.setColor(Color.lightGray);
         g.drawRect(0,0,300,300);
         //g.fillRect(0,0,300,300);
    	}
    }
}


class MImageFrame extends Frame {
	MImage i;
	
	public MImageFrame(MImage i) {
		this.i = i;
		final MImage output = i;
		
		//TODO: se le dimensione dell'immagine sono > dello schermo, aggiunge delle scrollbars?
		
		this.setSize(i.getWidth()+50, i.getHeight()+50);
		
		// aggiunge un pulsante per salvare l'immagine
		Button saveButton = new Button("Save...");
		saveButton.addActionListener( new ActionListener() {
			public void actionPerformed( ActionEvent e ) {
				FileDialog saveDialog =
					new FileDialog( HumanHeadDetectorGUI.mainFrame, "Save...", FileDialog.SAVE );
				saveDialog.setFilenameFilter( new FilenameFilter() {
					// MEMO: Filename filters do not function in Sun's reference implementation for Microsoft Windows.
					@Override
					public boolean accept(File directory, String filename) {
						if (filename.endsWith(".png")) return true;
						if (filename.endsWith(".gif")) return true;
						if (filename.endsWith(".jpg")) return true;
						if (filename.endsWith(".bmp")) return true;
						return false;
					}
				});
				saveDialog.setLocationRelativeTo(HumanHeadDetectorGUI.mainFrame);
				saveDialog.setVisible(true);
				
				if(saveDialog.getFile()==null) return; // l'utente ha premuto "Cancel"

				File outputFile = new File( saveDialog.getDirectory() + saveDialog.getFile() );
				
				try {
					output.save(outputFile);
				} catch (IOException e1) {
					new MessageDialog(HumanHeadDetectorGUI.mainFrame, "Error saving image", e1.getMessage(),true);
					return;
				}
			}
		});
		this.add(saveButton);
		this.setLayout(new FlowLayout());
		
		this.addWindowListener( new WindowAdapter() {
			public void windowClosing( WindowEvent e ) {
				dispose();
			}
		});
		
		this.setLocationRelativeTo(HumanHeadDetectorGUI.mainFrame);
		this.setVisible(true);
	}
	
    public void paint(Graphics g)
    {      
    	if(i!=null)
    		g.drawImage(i.i, 30, 30, this);
    }
	
}

// dialog che mostra un messaggio di testo statico
class MessageDialog extends Dialog implements ActionListener {

	private Label messageLabel;
	
	/**
	 * Crea un dialog contente un messaggio statico di testo e attende la conferma dell'utente.
	 * 
	 * @param f   Il ParentFrame a cui associare il dialog (NOTA: verrà bloccato se modal==true)
	 * @param title   Il titolo della finestra
	 * @param message    Il messaggio da mostrare all'utente all'interno della finestra
	 * @param modal   Se true blocca il parent frame in attesa che l'utente chiuda il dialog
	 */
	public MessageDialog(Frame parentFrame, String title, String message, boolean modal) {
		super(parentFrame, title, modal);
		//super((Frame)null, title, modal);
		
		messageLabel = new Label(message);
		this.add(messageLabel);
		
		Button okButton = new Button("Ok");
		okButton.addActionListener(this);
		this.add(okButton);

		this.setLayout(new FlowLayout(FlowLayout.CENTER));
		//this.setPreferredSize(new Dimension(messageLabel.getWidth(), messageLabel.getHeight() + okButton.getHeight()));
		this.pack();
		
		// centra la finestra rispetto al parent frame
		this.setLocationRelativeTo(parentFrame);
				
		this.setVisible(true);
	}
		
	public void actionPerformed(ActionEvent e) {
		dispose();
	}

}


class Stage1OptionsDialog extends Dialog implements ActionListener {
	
	boolean isConfirmed;
	
	boolean skinSegmentation;
	private int binaryThreshold;
	
	Label skinSegmentationLabel;
	Label binaryThresholdLabel;
	
	Checkbox skinSegmentationCheckbox;
	TextField binaryThresholdTextField;

	public Stage1OptionsDialog(Frame parentFrame) {
		super(parentFrame, "Preprocess options", true);
		
		// inizializza i componenti
		skinSegmentationLabel = new Label("segmentazione della pelle basata su colore");
		binaryThresholdLabel = new Label("threshold per binarizazione");
		
		skinSegmentationCheckbox = new Checkbox();
		skinSegmentationCheckbox.setState(true);
		
		binaryThresholdTextField = new TextField("10", 4);
		
		// aggiunge i componenti al dialog
		this.setLayout(new GridLayout(3,2,10,10));
		this.add(skinSegmentationLabel);
		this.add(skinSegmentationCheckbox);
		this.add(binaryThresholdLabel);
		this.add(binaryThresholdTextField);
		
		Button okButton = new Button("Ok");
		okButton.addActionListener(this);
		this.add(okButton);
		
		Button cancelButton = new Button("Cancel");
		cancelButton.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				isConfirmed = false;
				dispose(); // chiude la finestra
			}
		});
		this.add(cancelButton);
		
		this.pack();
		this.setLocationRelativeTo(parentFrame);
		this.setVisible(true);
	}
		
	public int getBinaryThreshold() {
		return(binaryThreshold);
	}

	@Override
	public void actionPerformed(ActionEvent e) {
		isConfirmed = true;
		
		skinSegmentation = skinSegmentationCheckbox.getState();
		
		try {
			binaryThreshold = Integer.parseInt(binaryThresholdTextField.getText());
			if(binaryThreshold<=0) throw(new Exception("threshold non valido"));
		} catch(Exception ex) {
			new MessageDialog(HumanHeadDetectorGUI.mainFrame, "Invalid value", ex.getMessage(), true);
			return; // mantiene aperta la finestra
		}
		
		dispose();
		
	}
}
	
class Stage2OptionsDialog extends Dialog implements ActionListener {
	
	boolean isConfirmed;
	
	private int minMajorAxis;
	private int maxMajorAxis;
	private int minMinorAxis;
	private int maxMinorAxis;
	private int noEllipses;
		
	//Label minMajorAxisLabel;
	//Label maxMajorAxisLabel;
	Label minMinorAxisLabel;
	Label maxMinorAxisLabel;
	Label noEllipsesLabel;
	
	//TextField minMajorAxisTextField;
	//TextField maxMajorAxisTextField;
	TextField minMinorAxisTextField;
	TextField maxMinorAxisTextField;
	TextField noEllipsesTextField;
    
	public Stage2OptionsDialog(Frame parentFrame) {
		super(parentFrame, "Ellipse detection options", true);
		
		//minMajorAxisLabel = new Label("semiasse maggiore min (px): ");
		//maxMajorAxisLabel = new Label("semiasse maggiore max (px): ");
		minMinorAxisLabel = new Label("semiasse minore min (px): ");
		maxMinorAxisLabel = new Label("semiasse minore max (px): ");
		noEllipsesLabel = new Label("numero di ellissi: ");
		
		//minMajorAxisTextField = new TextField("10", 4); 
		//maxMajorAxisTextField = new TextField("200", 4);
		minMinorAxisTextField = new TextField("10", 4); 
		maxMinorAxisTextField = new TextField("50", 4);
		noEllipsesTextField = new TextField("1", 3);
		
		this.setLayout(new GridLayout(4,2,10,10));
		/*
		this.add(minMajorAxisLabel);
		this.add(minMajorAxisTextField);
		this.add(maxMajorAxisLabel);
		this.add(maxMajorAxisTextField);
		*/
		this.add(minMinorAxisLabel);
		this.add(minMinorAxisTextField);
		this.add(maxMinorAxisLabel);
		this.add(maxMinorAxisTextField);
		this.add(noEllipsesLabel);
		this.add(noEllipsesTextField);
		
		Button okButton = new Button("Ok");
		okButton.addActionListener(this);
		this.add(okButton);
		
		Button cancelButton = new Button("Cancel");
		cancelButton.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				isConfirmed = false;
				dispose(); // chiude la finestra
			}
		});
		this.add(cancelButton);
		
		this.pack();
		this.setLocationRelativeTo(parentFrame);
		this.setVisible(true);
	}

	@Override
	public void actionPerformed(ActionEvent arg0) {
		isConfirmed = true;
		// legge i dati immessi dall'utente
		try {
			//minMajorAxis = Integer.parseInt(minMajorAxisTextField.getText());
			//maxMajorAxis = Integer.parseInt(maxMajorAxisTextField.getText()); 
			minMinorAxis = Integer.parseInt(minMinorAxisTextField.getText());
			if(minMinorAxis<=0) throw(new Exception("dimensione semiasse minore min non valida"));
			minMajorAxis = (int) (minMinorAxis*1.5);
			maxMinorAxis = Integer.parseInt(maxMinorAxisTextField.getText());
			if(maxMinorAxis<=0) throw(new Exception("dimensione semiasse minore max non valida"));
			maxMajorAxis = (int) (minMajorAxis*1.5); // maxMinorAxis?
			noEllipses = Integer.parseInt(noEllipsesTextField.getText());
			if(noEllipses<=0) throw(new Exception("numero di ellissi non valido"));
		} catch(Exception e) {
			new MessageDialog(HumanHeadDetectorGUI.mainFrame, "Invalid value", e.getMessage(), true);
			return; // mantiene aperta la finestra
		}
		
		dispose();
	}
	
	public int getMinMajorAxis() {
		return minMajorAxis;
	}
	public int getMaxMajorAxis() {
		return maxMajorAxis;
	}
	public int getMinMinorAxis() {
		return minMinorAxis;
	}
	public int getMaxMinorAxis() {
		return maxMinorAxis;
	}
	public int getNoEllipses() {
		return noEllipses;
	}

}
