import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.awt.image.BufferedImageOp;
import java.awt.image.ColorModel;

/**
 * 
 */

/**
 * @author Andrea
 *
 */
public class HumanSkinFilter implements BufferedImageOp {

	public Rectangle2D getBounds2D(BufferedImage src) {
	    return new Rectangle(0, 0, src.getWidth(),
	                         src.getHeight());
	  }

	  public BufferedImage createCompatibleDestImage(BufferedImage src, ColorModel destCM) {
	    if (destCM == null) {
	        destCM = src.getColorModel();
	    }

	    return new BufferedImage(destCM,
	        destCM.createCompatibleWritableRaster(
	          src.getWidth(), src.getHeight()),
	        destCM.isAlphaPremultiplied(), null);
	  }

	  public Point2D getPoint2D(Point2D srcPt,
	                            Point2D dstPt) {
	    return (Point2D) srcPt.clone();
	  }

	  public RenderingHints getRenderingHints() {
	    return null;
	  }

	@Override
	public BufferedImage filter(BufferedImage src, BufferedImage dest) {
		if(dest==null) {
			// create a new buffered image
			dest = new BufferedImage( src.getWidth(), 
					src.getHeight(),
					BufferedImage.TYPE_INT_RGB);
			
			// fill it with black
			Graphics2D g = dest.createGraphics();
			g.setColor(Color.BLACK);
			g.fill(new Rectangle(src.getWidth(), src.getHeight()));
		}
		
		for(int x=0;x<dest.getWidth();x++) {
			for(int y=0;y<dest.getHeight();y++) {
				
				// legge il colore corrente
				Color cRGB = new Color(src.getRGB(x, y));
				
				// converte il colore da RGB a HSB
				float[] cHSBcomponents = Color.RGBtoHSB(cRGB.getRed(),
														cRGB.getGreen(),
														cRGB.getBlue(),
														null);
				float hue = cHSBcomponents[0];
				int hueAngle = (int) ((cHSBcomponents[0] - Math.floor(cHSBcomponents[0]))*360);
				//System.out.println(hueAngle);
				float saturation = cHSBcomponents[1];
				float brightness = cHSBcomponents[2];
				
				// controlla i valori del colore corrente
				// se rispetta i range:
				// 0 < H < 50
				// 20% < S < 85%
				// B < 35%
				// MEMO: caucasian race only?
				if(
					(hueAngle>=0 && hueAngle<=50) &&
					(saturation>=0.2 && saturation<=0.85) &&
					(brightness>=0.35)
				   )
					dest.setRGB(x, y, src.getRGB(x, y)); // lo copia nell'immagine di destinazione
				// altrimenti lo scarta (il pixel resta nero)
			}
		}
		
		return(dest);
	}


}
