import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.Stroke;
import java.awt.geom.AffineTransform;
import java.awt.image.*;
import java.io.File;
import java.io.IOException;
import javax.imageio.ImageIO;


public class MImage /*extends BufferedImage*/ implements Cloneable { 

	public BufferedImage i; // TODO: provare ad evitare embedding?
	public Image getImage() {
		return(i);
	}
	
	/**
	 * Create an empty image filled with a solid color.
	 * @param width
	 * @param height
	 * @param imageType
	 */
	public MImage(int width, int height, int imageType, Color c) {
		i = new BufferedImage(width, height, imageType);		
		if(c!=null && c!=Color.BLACK)
			this.fill(new Rectangle(width, height), c);
	}
	
	/**
	 * Load an image from a graphic file.
	 * Supported formats are: GIF, PNG, JPEG, BMP (using javax.imageio.ImageIO)
	 * @param f
	 * @throws IOException 
	 */
	public MImage(File f) throws IOException {
		if(f==null) return;
		i = ImageIO.read( f );
		if(i==null) throw new IOException();
		System.out.println("image file \"" + f + "\" load successfully.");
		System.out.println(this.toString());
	}
	
	/* (non-Javadoc)
	 * @see java.lang.Object#clone()
	 */
	@Override
	protected Object clone() throws CloneNotSupportedException {
		return super.clone();
	}
	
	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		String[] propertyNames = i.getPropertyNames();
		if( propertyNames != null ) 
		{
			String properties = "";
			for( String p : propertyNames ) {
				properties += p + ": " + i.getProperty(p) + '\n';
			}
			return properties;
		}
		else
			return(
					"Width: " + i.getWidth() + '\n' +
					"Height: " + i.getHeight() + '\n' +
					"ColorModel: " + i.getColorModel().toString() + '\n' +
					"SampleModel: " + i.getSampleModel().toString()
					//...
					);
	}

	/**
	 * Save current image into a file.
	 * Supported formats are: GIF, PNG, JPEG, BMP (using javax.imageio.ImageIO)
	 * @throws IOException 
	 */
	public void save(File f) throws IOException {
		if(f==null) return;
		ImageIO.write( i, findOutputFileFormat(f.toString()), f );
		System.out.println("image file \"" + f + "\" saved successfully.");
	}
	private String findOutputFileFormat(String s) { // TODO: put inside save method
		int lastDotPos = s.lastIndexOf('.');
		if( lastDotPos == -1 || lastDotPos+1 >= s.length() )
			return("png"); // if no extension specified, assume PNG as default
		else
			return s.substring( lastDotPos + 1 );
	}
	
	public int getWidth() {
		return i.getWidth();
	}
	
	public int getHeight() {
		return i.getHeight();
	}
	
	/**
	 * Read the pixel color value at the specified cords.
	 * @param x
	 * @param y
	 * @return
	 */
	public Color getPixel(int x, int y) {
		if(x>i.getWidth() || y>i.getHeight()) return null; // bounds checking
		int c = i.getRGB(x, y);
		return new Color(c, true);
	}
	
	public void setPixel(int x, int y, Color c) {
		if(x>i.getWidth() || y>i.getHeight()) return; // bounds checking
		if(c==null) c = Color.WHITE; // default argument
		i.setRGB(x, y, c.getRGB());
	}
	
	public void draw(Shape s, Color c, Stroke st) {
		Graphics2D g = i.createGraphics();
		if(c!=null) g.setColor(c);
		if(st!=null) g.setStroke(st);
		g.draw(s);
	}
	
	public void fill(Shape s, Color c) {
		Graphics2D g = i.createGraphics();
		if(c!=null) g.setColor(c);
		g.fill(s);
	}
	
	public void filter( BufferedImageOp op ) {
		i = op.filter( i, null );
	}
	
	//TODO? public int[][] getGradient();
	
	public int[][] getHistogram() {	
		int[] frequenciesRed = new int[256];
		int[] frequenciesGreen = new int[256];
		int[] frequenciesBlue = new int[256];
		// inizializzazione a 0 automatica?
		
		for(int y=0; y<i.getHeight(); y++)
			for(int x=0; x<i.getWidth(); x++)
			{
				Color p = getPixel(x, y);
				
				frequenciesRed[p.getRed()]++;
				frequenciesGreen[p.getGreen()]++;
				frequenciesBlue[p.getBlue()]++;
			}
		
		// print frequencies
		for(int i=0; i<256; i++)
			System.out.println("intensity: " + i + " - frequencies: red=" + frequenciesRed[i] + " green=" + frequenciesGreen[i] + " blue=" + frequenciesBlue[i] );
				
		int[][] frequencies = { frequenciesRed, frequenciesGreen, frequenciesBlue };
		return(frequencies);
	}

	
	// TODO? normalize()
	// TODO? equalize()
	
	
//*************************** COLOR CONVERSION FILTERS

	public void toGrayscale() {
		BufferedImage newI = new BufferedImage(i.getWidth(), i.getHeight(), BufferedImage.TYPE_BYTE_GRAY);
		Graphics g = newI.getGraphics();
		g.drawImage(i, 0, 0, null);
		g.dispose();
		i = newI;
		//ALTERNATIVE: filter( new ColorConvertOp( ColorSpace.getInstance( ColorSpace.CS_GRAY ), null ) ); 
	}
	
	public void toRGB() {
		BufferedImage newI = new BufferedImage(i.getWidth(), i.getHeight(), BufferedImage.TYPE_INT_ARGB);
		Graphics g = newI.getGraphics();
		g.drawImage(i, 0, 0, null);
		g.dispose();
		i = newI;
		//ALTERNATIVE: filter( new ColorConvertOp( ColorSpace.getInstance( ColorSpace.CS_sRGB ), null ) );
	}
	
	public void toBinary() {
		BufferedImage newI = new BufferedImage(i.getWidth(), i.getHeight(), BufferedImage.TYPE_BYTE_BINARY);
		Graphics g = newI.getGraphics();
		g.drawImage(i, 0, 0, null);
		g.dispose();
		i = newI;
	}
	
	public void toIndexed () { // convert to 256-color indexed palette
		BufferedImage newI = new BufferedImage(i.getWidth(), i.getHeight(), BufferedImage.TYPE_BYTE_INDEXED);
		Graphics g = newI.getGraphics();
		g.drawImage(i, 0, 0, null);
		g.dispose();
		i = newI;
	}
	

// ********************* GEOMETRIC TRANSFORMATIONS
	
	public void rotate( int degrees ) {
		if(degrees>=360) degrees-=360;
		AffineTransform tx = AffineTransform.getRotateInstance( Math.toRadians( degrees ) );
		if(degrees==90) tx.translate( 0, i.getHeight() * -1 );
		if(degrees==180) tx.translate( i.getWidth() * -1, i.getHeight() * -1 );
		if(degrees==270) tx.translate( i.getWidth() * -1, 0 );
		
		filter( new AffineTransformOp( tx, AffineTransformOp.TYPE_BILINEAR ) );
	}
	   
	public void scale(double factor) {
		AffineTransform tx = new AffineTransform();
		tx.scale( factor, factor );

		filter( new AffineTransformOp( tx, AffineTransformOp.TYPE_BILINEAR ) );
	}

	public void scale(int newWidth, int newHeight) {
		AffineTransform tx = new AffineTransform();
		tx.scale( (newWidth /(double) i.getWidth()), (newHeight /(double) i.getHeight()) );
		
		filter( new AffineTransformOp( tx, AffineTransformOp.TYPE_BILINEAR ) );
	}
	
	public void flipVertical() {
		AffineTransform tx = AffineTransform.getScaleInstance(1, -1);
	    tx.translate(0, -1 * i.getHeight());
	    
	    filter( new AffineTransformOp(tx, AffineTransformOp.TYPE_BILINEAR ) );
	}
	
	public void flipHorizontal() {
		AffineTransform tx = AffineTransform.getScaleInstance(-1, 1);
		tx.translate(-1 * i.getWidth(), 0);
		
		filter( new AffineTransformOp(tx, AffineTransformOp.TYPE_BILINEAR ) );
	}

	
//***********************  IMAGE Arithmetic
	
	public MImage sum(MImage addendum) {
		MImage output = new MImage( Math.max(this.getWidth(), addendum.getWidth()), 
				                    Math.max(this.getHeight(), addendum.getHeight()),
				                    BufferedImage.TYPE_INT_RGB,
				                    Color.BLACK );

		for(int x=0;x<output.getWidth();x++) {
			for(int y=0;y<output.getHeight();y++) {
				/* wrong?
				int a1 = this.getPixel(x, y).getRGB();
				int a2 = addendum.getPixel(x, y).getRGB();
				output.setPixel(x, y, new Color(a1+a2));
				*/
				int r = this.getPixel(x, y).getRed() + addendum.getPixel(x, y).getRed();
				if(r>255) r = 255;
				int g = this.getPixel(x, y).getGreen() + addendum.getPixel(x, y).getGreen();
				if(g>255) g = 255;
				int b = this.getPixel(x, y).getBlue() + addendum.getPixel(x, y).getBlue();
				if(b>255) b = 255;
				output.setPixel(x, y, new Color(r,g,b));
			}
		}
		
		return(output);
	}
	
	// TODO: subtract
	
	
	
//************************* FILTERS
	
	public void threshold(int threshold) {	
		byte[] thresholdArray = new byte[256];
		for (int i = 0; i < 256; i++) {
			if (i < threshold)
				thresholdArray[i] = (byte)0;
			else
				thresholdArray[i] = (byte)255;
		}
		
		filter( new LookupOp(new ByteLookupTable( 0, thresholdArray ), null) );
	}
	
	public void negate() {
		byte[] invert = new byte[256];
		for (int i = 0; i < 256; i++)
			invert[i] = (byte)(255 - i);
		filter( new LookupOp( new ByteLookupTable( 0, invert ), null ) );
	}
	
	public void uniformBlur() {  // aka "Mean Filter"
		float ninth = 1.0f / 9.0f;
		float[] blurKernel = {
			ninth, ninth, ninth,
			ninth, ninth, ninth,
			ninth, ninth, ninth
		};
				
		filter( new ConvolveOp( new Kernel( 3, 3, blurKernel ) ) );
	}
	
	public void gaussianBlur() {
		float[] gaussianKernel = { // Discrete approximation to Gaussian function with sigma=1.0 
			1, 4, 7, 4, 1,
			4,16,26,16, 4,
			7,26,41,26, 7,
			4,16,26,16, 4,
			1, 4, 7, 4, 1
		};
		for(int i=0; i<gaussianKernel.length; i++)
			gaussianKernel[i] /= 273.0f;
		
		filter( new ConvolveOp( new Kernel( 5, 5, gaussianKernel ) ) );
	}
	
	/* TODO?
	public void MedianFilter() {
		
	}*/
	
	public void brighten(float factor) // factor range: >1.0 = brighten, <1.0 darken
	{
		filter( new RescaleOp(factor, 0, null) );
	}
	
	public void darken(float factor) // factor range: >1.0 = brighten, <1.0 darken
	{
		filter( new RescaleOp(factor, 0, null) );
	}

	/* edge detector using Sobel kernel (1st order derivative) */
	public void edge() {
		MImage c1 = null;
		MImage c2 = null;
		//MImage c3 = null;
		//MImage c4 = null;
		try {
			c1 = (MImage) this.clone();
			c2 = (MImage) this.clone();
			//c3 = (MImage) this.clone();
			//c4 = (MImage) this.clone();
		} catch (CloneNotSupportedException e) {
		}

		// crea 2 cloni dell'immagine di partenza, applica i 2 kernel e somma i risultati
		
		float[] SOBEL_V1 = {
			-1.0F, -2.0F, -1.0F,
			 0.0F,  0.0F,  0.0F,
			 1.0F,  2.0F,  1.0F
			};
				
		float[] SOBEL_V2 = {
				 1.0F,  2.0F,  1.0F,
				 0.0F,  0.0F,  0.0F,
				-1.0F, -2.0F, -1.0F
			};
		
		float[] SOBEL_H2 = {
				 1.0F,  0.0F, -1.0F,
				 2.0F,  0.0F, -2.0F,
				 1.0F,  0.0F, -1.0F
				};
		
		float[] SOBEL_H1 = {
				-1.0F,  0.0F,  1.0F,
				-2.0F,  0.0F,  2.0F,
				-1.0F,  0.0F,  1.0F
				};
		
		c1.filter ( new ConvolveOp( new Kernel( 3, 3, SOBEL_V1 ) ) );
		//c2.filter ( new ConvolveOp( new Kernel( 3, 3, SOBEL_V2 ) ) );
		c2.filter ( new ConvolveOp( new Kernel( 3, 3, SOBEL_H1 ) ) );
		//c4.filter ( new ConvolveOp( new Kernel( 3, 3, SOBEL_H2 ) ) );
		
		i = c1.sum(c2).i;
		//i = c1.sum(c2).sum(c3).sum(c4).i;
	}
	
	/* edge detector using laplacian (2nd order derivative) */
	public void edge2() {
		float[] laplacianKernel = {
				-1.0f, -1.0f, -1.0f,
				-1.0f,  8.0f, -1.0f,
				-1.0f, -1.0f, -1.0f
			};
		
		filter ( new ConvolveOp( new Kernel( 3, 3, laplacianKernel ) ) );
	}
	
	/** sharpen using laplacian "High-boost" filter
	 * @param factor
	 */
	public void sharpen(float boost) {
		float[] laplacianKernel = {
				-1.0f, -1.0f, -1.0f,
				-1.0f,  (boost+9.0f), -1.0f,
				-1.0f, -1.0f, -1.0f
			};
		
		filter ( new ConvolveOp( new Kernel( 3, 3, laplacianKernel ) ) );
	}
	public void sharpen() {
		sharpen(0.0f);
	}
	
}
